import copy
import logging
import os.path
import types
from abc import ABC, abstractmethod
from collections import OrderedDict
from collections.abc import Iterable

import mrcfile
import numpy as np

from aspire.abinitio import CLOrient3D, CLSyncVoting
from aspire.image import Image, normalize_bg
from aspire.image.xform import (
    Downsample,
    FilterXform,
    IndexedXform,
    LambdaXform,
    Multiply,
    Pipeline,
)
from aspire.noise import NoiseEstimator, WhiteNoiseEstimator
from aspire.operators import (
    CTFFilter,
    Filter,
    IdentityFilter,
    MultiplicativeFilter,
    PowerFilter,
)
from aspire.storage import MrcStats, StarFile
from aspire.utils import Rotation, grid_2d, support_mask, trange
from aspire.volume import IdentitySymmetryGroup, SymmetryGroup

logger = logging.getLogger(__name__)


class _ImageAccessor:
    """
    Helper class for accessing images from an ImageSource as slices via the `src.images[start:stop:step]` API.
    """

    def __init__(self, fun, num_imgs):
        """
        :param fun: The private image-accessing method specific to the ImageSource associated with this ImageAccessor.
                    Generally _images() but can be substituted with a custom method.
        :param num_imgs: The max number of images that this ImageAccessor can load (generally ImageSource.n).
        """
        self.fun = fun
        self.num_imgs = num_imgs

    def __getitem__(self, indices):
        """
        ImageAccessor can be indexed via Python slice object, 1-D NumPy array, list, or a single integer,
        corresponding to the indices of the requested images. By default, slices default to a start of 0,
        an end of self.num_imgs, and a step of 1.
        :return: An Image object containing the requested images.
        """
        if isinstance(indices, Iterable) and not isinstance(indices, np.ndarray):
            indices = np.fromiter(indices, int)
        elif isinstance(indices, (int, np.integer)):
            indices = np.array([indices])
        elif isinstance(indices, slice):
            start, stop, step = indices.start, indices.stop, indices.step

            # X[:end], slice(None, e, None) -> slice(0, e, 1)
            if not start:
                start = 0

            # Special case for X[-10:]
            if start < 0 and stop is None:
                # slice(-10, None, None) -> slice(-10, *0* ,1)
                stop = 0
            # All other cases, limit to num_imgs
            #   slice(s, None, None) -> slice(s, num_imgs, 1)
            #   slice(s, 10**10, None) -> slice(0, num_imgs, 1)
            elif not stop or stop > self.num_imgs:
                stop = self.num_imgs

            if not step:
                step = 1

            if not all(isinstance(i, (int, np.integer)) for i in [start, stop, step]):
                raise TypeError("Non-integer slice components.")
            indices = np.arange(start, stop, step)

        if not isinstance(indices, np.ndarray):
            raise KeyError(
                "Key for .images must be a slice, 1-D NumPy array, or iterable yielding integers."
            )
        if not indices.ndim == 1:
            raise KeyError("Only one-dimensional indexing is allowed for images.")

        # final check for out-of-range indices
        out_of_range = indices >= self.num_imgs
        if out_of_range.any():
            raise KeyError(f"Out-of-range indices: {list(indices[out_of_range])}")

        # check for negative indices and flip to positive
        indices = indices % self.num_imgs

        return self.fun(indices)


def as_copy(func):
    """
    Method decorator that invokes the decorated method on a deepcopy of the object,
    and returns it on exit. The original object is unmodified.
    This allows one to take a mutating method on an object:
       obj.increment(by=2)
    and use it in a functional way:
       another = obj.increment(by=2)  # obj unmodified
    Note that the original return value of the method is lost, so this decorator
    is best used on methods that mutate the object but don't return anything.
    """

    def wrapper(self, *args, **kwargs):
        obj_copy = copy.deepcopy(self)
        func_copy = copy.deepcopy(func)
        func_copy(obj_copy, *args, **kwargs)
        return obj_copy

    return wrapper


class ImageSource(ABC):
    """
    When creating an `ImageSource` object, a 'metadata' table holds metadata information about all images in the
    `ImageSource`. The number of rows in this metadata table will equal the total number of images supported by this
    `ImageSource` (available as the 'n' attribute), though reading/writing of images is usually done in chunks.

    This metadata table is implemented as a dictionary of numpy arrays.

    The 'values' in this metadata table are usually primitive types (floats/ints/strings) that are suitable
    for being read from STAR files, and being written to STAR files. The columns corresponding to these fields
    begin with a single underscore '_'.

    In addition, the metadata table may also contain references to Python objects.
    `Filter` objects, for example, are stored in this metadata table as references to unique `Filter` objects that
    correspond to images in this `ImageSource`. Several rows of metadata may end up containing a reference to a small
    handful of unique `Filter` objects, depending on the values found in other columns (identical `Filter`
    objects). For example, a smaller number of CTFFilter objects may apply to subsets of particles depending on
    the unique "_rlnDefocusU"/"_rlnDefocusV" Relion parameters.
    """

    # The abstract class starts off _mutable. Conrete classes should
    # disable _mutable as the last step in __init__.
    _mutable = True

    def __init__(
        self, L, n, dtype="double", metadata=None, memory=None, symmetry_group=None
    ):
        """
        A cryo-EM ImageSource object that supplies images along with other parameters for image manipulation.

        :param L: resolution of (square) images (int)
        :param n: The total number of images available
            Note that images() may return a different number of images based on its arguments.
        :param metadata: A Dataframe of metadata information corresponding to this ImageSource's images
        :param memory: str or None
            The path of the base directory to use as a data store or None. If None is given, no caching is performed.
        :param symmetry_group: A SymmetryGroup instance or string indicating the underlying symmetry of the molecule.
            Defaults to the `IdentitySymmetryGroup`, which represents an asymmetric particle, if none provided.
        """

        # Instantiate the accessor for the `images` property
        self._img_accessor = _ImageAccessor(self._images, n)

        self.L = L
        self._n = None
        self.n = n
        self.dtype = np.dtype(dtype)

        # The private attribute '_cached_im' can be populated by calling this object's cache() method explicitly
        self._cached_im = None

        # _rotations is assigned non None value
        #  by `rotations` or `angles` setters.
        #  It is potentially used by subclasses to test if we've used setters.
        #  This must come before the Relion/starfile metadata parsing below.
        self._rotations = None

        if metadata is None:
            self._metadata = {}
        else:
            self._metadata = copy.copy(metadata)
            if self.has_metadata(["_rlnAngleRot", "_rlnAngleTilt", "_rlnAnglePsi"]):
                self._rotations = Rotation.from_euler(
                    np.deg2rad(
                        self.get_metadata(
                            ["_rlnAngleRot", "_rlnAngleTilt", "_rlnAnglePsi"]
                        )
                    )
                )

        self._populate_symmetry_group(symmetry_group)

        self.unique_filters = []
        self.generation_pipeline = Pipeline(xforms=None, memory=memory)

        logger.info(f"Creating {self.__class__.__name__} with {len(self)} images.")

    def __deepcopy__(self, memo):
        """
        A custom __deepcopy__ implementation to individually handle special cases.
        Mostly copied over from https://stackoverflow.com/a/71125311
        """
        # Get a reference to the bound deepcopy method
        deepcopy_method = self.__deepcopy__
        # Temporarily disable __deepcopy__ to avoid infinite recursion
        self.__deepcopy__ = None
        # Create a deepcopy cp using the normal procedure
        cp = copy.deepcopy(self, memo)

        # --------------------------------------
        # Handle any special cases for cp here.
        # --------------------------------------
        # This is the whole reason this method exists. If this section is empty,
        # then this entire __deepcopy__ implementation can be removed.

        # The 'dtype' attribute is a numpy module level singleton obtained by np.dtype(..) call
        # The 'finufft' library currently compares this to the result of a new np.dtype(..) call
        # by reference, not by value (as it should). A deepcopy will make a copy of the singleton,
        # and thus comparison by reference will fail. Till this bug in 'finufft' is removed, we assign
        # self.dtype to dtype
        cp.dtype = self.dtype

        # --------------------------------------

        # Reattach the bound deepcopy method
        self.__deepcopy__ = deepcopy_method
        # Get the unbounded function corresponding to the bound deepcopy method and rebind to cp
        cp.__deepcopy__ = types.MethodType(deepcopy_method.__func__, cp)

        return cp

    @property
    def symmetry_group(self):
        """
        A SymmetryGroup instance associated with the symmetry type of the ImageSource object.
        Access rotation matrices of the symmetry_group via `symmetry_group.matrices`.
        """
        return self._symmetry_group

    @symmetry_group.setter
    def symmetry_group(self, value):
        """
        Set the `symmetry_group` for `src`.

        :param value: A `SymmetryGroup` instance or string indicating symmetry, ie. "C5", "D7", "T", etc.
        """
        if not self._mutable:
            raise RuntimeError(
                f"This source is no longer mutable. Try new_source = source.update(symmetry_group='{value}')."
            )
        if isinstance(value, str):
            value = SymmetryGroup.from_string(value, dtype=self.dtype)
        if not isinstance(value, SymmetryGroup):
            raise ValueError(
                "`symmetry_group` must be an instance of the SymmetryGroup class"
            )
        self._symmetry_group = value
        self.set_metadata(["_rlnSymmetryGroup"], str(self.symmetry_group))

    def _populate_symmetry_group(self, symmetry_group):
        """
        Populates the symmetry_group attribute with user provided symmetry_group or metadata.
        If neither exist, defaults to C1 symmetry.
        """
        if self.has_metadata(["_rlnSymmetryGroup"]):
            if symmetry_group:
                logger.warning(
                    f"Overriding metadata with supplied symmetry group {symmetry_group}"
                )
            else:
                symmetry_group = SymmetryGroup.from_string(
                    symmetry=self.get_metadata(["_rlnSymmetryGroup"])[0],
                    dtype=self.dtype,
                )

        self.symmetry_group = symmetry_group or IdentitySymmetryGroup(dtype=self.dtype)

    def __getitem__(self, indices):
        """
        Check `indices` and return slice of current Source as a new
        Source.

        Internally uses `IndexedSource`.

        :param indices: Requested indices as a Python slice object,
            1-D NumPy array, list, or a single integer. Slices default
            to a start of 0, an end of self.num_imgs, and a step of 1.
            See _ImageAccessor.
        :return: Source composed of the images and metadata at `indices`.
        """

        return IndexedSource(self, indices)

    def __len__(self):
        """
        Returns total number of images in source.
        """
        return self.n

    def _metadata_as_dict(self, metadata_fields, indices, default_value=None):
        """
        Return a dictionary of selected metadata fields at selected indices.
        :param metadata_fields: An iterable of strings specifying metadata fields.
        :param indices: An ndarray of 0-indexed locations we're interested in.
        :param default_value: A scalar default value to use if a metadata_field is not found.
        :return: A dictionary of numpy arrays of specified metadata fields at specified indices.
        """
        result = {}
        for metadata_field in metadata_fields:
            if metadata_field in self._metadata:
                result[metadata_field] = self._metadata[metadata_field][indices].copy()
            else:
                if default_value is None:
                    raise ValueError(
                        f"Missing metadata field {metadata_field} and no default_value supplied"
                    )
                result[metadata_field] = np.full(len(indices), fill_value=default_value)
        return result

    def _metadata_as_ndarray(self, metadata_fields, indices, default_value=None):
        """
        Return a numpy array of selected metadata fields at selected indices.
        :param metadata_fields: An iterable of strings specifying metadata fields.
        :param indices: An ndarray of 0-indexed locations we're interested in.
        :param default_value: A scalar default value to use if a metadata_field is not found.
        :return: A numpy array of specified metadata fields at specified indices.
        """
        # Start with the most generic type - we'll narrow it later
        result = np.empty((len(indices), len(metadata_fields))).astype("object")
        # Keep track of dtypes of individual metadata fields, so we can narrow the result into a single dtype
        dtypes = [None] * len(metadata_fields)

        for i, metadata_field in enumerate(metadata_fields):
            if metadata_field not in self._metadata:
                if default_value is None:
                    raise ValueError(
                        f"Missing metadata field {metadata_field} and no default_value supplied"
                    )
                result[:, i] = default_value
                dtypes[i] = np.array([default_value]).dtype
            else:
                values = self._metadata[metadata_field][indices]
                result[:, i] = values
                dtypes[i] = values.dtype

        dtype = np.result_type(*dtypes)
        result = result.astype(dtype)

        if result.shape[1] == 1:
            result = result.squeeze(axis=1)

        return result

    def update(self, **kwargs):
        """
        Update certain properties that modify the underlying metadata, and return a new ImageSource
        object with the new properties. The original object is unchanged.
        """
        updateable_props = (
            "states",
            "filter_indices",
            "offsets",
            "amplitudes",
            "angles",
            "rotations",
            "symmetry_group",
        )

        cp = copy.deepcopy(self)
        cp._mutable = True
        for prop in updateable_props:
            if prop in kwargs:
                setattr(cp, prop, kwargs.pop(prop))

        if kwargs:
            logger.warning(f"Unhandled arguments = {kwargs.keys()}")
        cp._mutable = False

        return cp

    @property
    def n(self):
        return self._n

    @n.setter
    def n(self, n):
        """
        Sets max image index `n` in `src` and associated
        `ImageAccessor`.

        :param n: Number of images.
        """
        self._set_n(n)

    def _set_n(self, n):
        """
        Sets max image index `n` in `src` and associated
        `ImageAccessor`.

        :param n: Number of images.
        """
        # Protect _n by default.
        if self._n is not None:
            raise RuntimeError("Source `n` is already set.")

        # Enforce type, just-in-case.
        if n != int(n):
            raise TypeError("`n` must be an integer")
        n = int(n)

        self._img_accessor.num_imgs = n
        self._n = n

    @property
    def n_ctf_filters(self):
        """
        Return the number of CTFFilters found in this Source.
        """
        return len([f for f in self.unique_filters if isinstance(f, CTFFilter)])

    @property
    def states(self):
        return np.atleast_1d(self.get_metadata("_rlnClassNumber"))

    @states.setter
    def states(self, values):
        return self.set_metadata("_rlnClassNumber", values)

    @property
    def filter_indices(self):
        return np.atleast_1d(self.get_metadata("__filter_indices"))

    @filter_indices.setter
    def filter_indices(self, indices):
        # create metadata of filters for all images
        return self.set_metadata(["__filter_indices"], indices)

    @property
    def offsets(self):
        return np.atleast_2d(
            self.get_metadata(["_rlnOriginX", "_rlnOriginY"], default_value=0.0)
        )

    @offsets.setter
    def offsets(self, values):
        return self.set_metadata(
            ["_rlnOriginX", "_rlnOriginY"], np.array(values, dtype=self.dtype)
        )

    @property
    def amplitudes(self):
        return np.atleast_1d(self.get_metadata("_rlnAmplitude", default_value=1.0))

    @amplitudes.setter
    def amplitudes(self, values):
        return self.set_metadata("_rlnAmplitude", np.array(values, dtype=self.dtype))

    @property
    def angles(self):
        """
        Rotation angles in radians.

        :return: Rotation angles in radians, as a n x 3 array
        """
        # Call a private method. This allows subclasses to efficiently override.
        return self._angles()

    def _angles(self):
        """
        Converts internal _rotations representation to expected matrix form.
        """
        return self._rotations.angles.astype(self.dtype)

    @property
    def rotations(self):
        """
        Returns rotation matrices.

        :return: Rotation matrices as a n x 3 x 3 array
        """
        # Call a private method. This allows sub classes to effeciently override.
        return self._rots()

    def _rots(self):
        """
        Converts internal `_rotations` representation to expected matrix form.

        :return: Rotation matrices as a n x 3 x 3 array
        """
        return self._rotations.matrices.astype(self.dtype)

    @angles.setter
    def angles(self, values):
        """
        Set rotation angles

        :param values: Rotation angles in radians, as a n x 3 array
        :return: None
        """

        values = values.astype(self.dtype)
        self._rotations = Rotation.from_euler(values)
        self.set_metadata(
            ["_rlnAngleRot", "_rlnAngleTilt", "_rlnAnglePsi"], np.rad2deg(values)
        )

    @rotations.setter
    def rotations(self, values):
        """
        Set rotation matrices

        :param values: Rotation matrices as a n x 3 x 3 array
        :return: None
        """

        values = values.astype(self.dtype)
        self._rotations = Rotation.from_matrix(values)
        self.set_metadata(
            ["_rlnAngleRot", "_rlnAngleTilt", "_rlnAnglePsi"],
            np.rad2deg(self._rotations.angles),
        )

    @property
    def class_indices(self):
        """
        Returns table of class image indices as `(src.n, n_nbors)`
        Numpy array.

        Each row reprsents a class, with the columns ordered by
        smallest `class_distances` from the reference image (zeroth
        columm).

        Note `n_nbors` is managed by `self.classifier` and used here
        for documentation.

        :return: Numpy array, integers.
        """
        res = self.get_metadata(["_class_indices"])
        return np.vstack([np.array(row.split(","), dtype=int) for row in res])

    @property
    def selection_indices(self):
        return self.get_metadata(["_selection_indices"])

    @property
    def class_refl(self):
        """
        Returns table of class image reflections as `(src.n, n_nbors)`
        Numpy array.

        Follows same layout as `class_indices` but holds booleans that
        are True when the image should be reflected before averaging.

        Note `n_nbors` is managed by `self.classifier` and used here
        for documentation.

        :return: Numpy array, boolean.
        """
        res = self.get_metadata(["_class_refl"])
        # Read table of (0, 1) integers, cast to `bool`.
        return np.vstack([np.array(row.split(","), dtype=int) for row in res]).astype(
            bool
        )

    @property
    def class_distances(self):
        """
        Returns table of class image distances as `(src.n, n_nbors)`
        Numpy array.

        Follows same layout as `class_indices` but holds floats
        representing the distance (returned by classifier) to the
        zeroth image in each class.

        Note `n_nbors` is managed by `self.classifier` and used here
        for documentation.

        :return: Numpy array, self.dtype.
        """
        res = self.get_metadata(["_class_distances"])
        return np.vstack([np.array(row.split(","), dtype=self.dtype) for row in res])

    def set_metadata(self, metadata_fields, values, indices=None):
        """
        Modify metadata field information of this ImageSource for selected indices

        :param metadata_fields: A string, or list of strings, representing the metadata field(s) to be modified
        :param values: A scalar or vector (of length |indices|) of replacement values.
        :param indices: A list of 0-based indices indicating the indices for which to modify metadata.
            If indices is None, then all indices in this Source object are modified. In this case,
            values should either be a scalar or a vector of length equal to the total number of images, |self.n|.
        :return: On return, the metadata associated with the specified indices has been modified.
        """

        # This breaks lots of things, maybe not something we want to rush out.
        # # Check if we're in an immutable state.
        # if not self._mutable:
        #     raise RuntimeError("This source is no longer mutable, try using `update` instead of `set_metadata`")

        if isinstance(metadata_fields, str):
            metadata_fields = [metadata_fields]

        if indices is None:
            indices = np.arange(self.n)

        # Check if we're an iterable, and in case we're not, broadcast
        # the single values into a list `indices` long.
        try:
            iter(values)
        except TypeError:
            values = [values] * len(indices)
        else:
            # Special case for single `str`, which are iterable, but
            # need to be broadcast like a singleton.
            if isinstance(values, str):
                values = [values] * len(indices)
        if len(values) != len(indices):
            raise RuntimeError(
                f"Mismatch between len(values) {len(values)} and len(indices) {len(indices)}."
            )

        values = np.array(values)  # make a copy for our use

        # When creating metadata fields that are string, coerce them into python objects to allow string expansion
        # later (replacing 'hello' with 'goodbye', for example).
        # This is in part to conform to legacy implementation of metadata that used pandas,
        # but is probably a good thing to do anyway.
        # We also come up with a sensible fill value while we're at it.
        if np.issubdtype(values.dtype, np.str_):
            values = values.astype("object")
            fill_value = ""
        elif np.issubdtype(values.dtype, np.integer):
            # For integers, we'll use the minimal value.
            # This will be a large negative value when signed,
            # and zero for unsigned integers.
            fill_value = np.iinfo(values.dtype).min
        else:
            fill_value = np.nan

        if values.ndim == 1:
            values = values.reshape(-1, 1)  # convert to column
            values = np.tile(
                values, (1, len(metadata_fields))
            )  # stack columns for each metadata field

        for i, metadata_field in enumerate(metadata_fields):
            if metadata_field not in self._metadata:
                self._metadata[metadata_field] = np.full(self.n, fill_value).astype(
                    values.dtype
                )
            self._metadata[metadata_field][indices] = values[:, i]

    def has_metadata(self, metadata_fields):
        """
        Find out if one or more metadata fields are available for this `ImageSource`.

        :param metadata_fields: A string, of list of strings, representing the metadata field(s) to be queried.
        :return: Boolean value indicating whether the field(s) are available.
        """
        if isinstance(metadata_fields, str):
            metadata_fields = [metadata_fields]
        return all(f in self._metadata for f in metadata_fields)

    def get_metadata(
        self, metadata_fields=None, indices=None, default_value=None, as_dict=False
    ):
        """
        Get metadata field information of this ImageSource for a
        selection of fields of indices.  The default should return the
        entire metadata table.

        :param metadata_fields: A string, or list of strings,
            representing the metadata field(s) to be queried.
            Defaults to None, which yields all populated columns.
        :param indices: A list of 0-based indices indicating the
            indices for which to get metadata.  If indices is None,
            then values corresponding to all indices in this Source
            object are returned.
        :param default_value: Default scalar value to use for any
            fields not found in the metadata. If None, no default
            value is used, and missing field(s) cause a RuntimeError.
        :param as_dict: Boolean indicating whether we want to return
            metadata as a dictionary (True), or as a numpy ndarray (False).
            In the latter case, all returned values are typecast to a common
            numpy dtype, so use with caution.
        :return: An ndarray of values (any valid np types)
            representing metadata info. If is_dict is True, then returns
            a dictionary mapping metadata names to numpy arrays of values.
        """
        if metadata_fields is None:
            metadata_fields = list(self._metadata.keys())
        if isinstance(metadata_fields, str):
            metadata_fields = [metadata_fields]

        if indices is None:
            indices = np.arange(self.n)
        else:
            try:
                iter(indices)
            except TypeError:
                indices = [indices]

        if as_dict:
            return self._metadata_as_dict(
                metadata_fields=metadata_fields,
                indices=indices,
                default_value=default_value,
            )
        else:
            return self._metadata_as_ndarray(
                metadata_fields=metadata_fields,
                indices=indices,
                default_value=default_value,
            )

    def _apply_filters(
        self,
        im_orig,
        filters,
        indices,
    ):
        """
        For images in `im_orig`, `filters` associated with the corresponding
        index in the supplied `indices` are applied. The images are then returned as an `Image` stack.

        :param im_orig: An `Image` object
        :param filters: A list of `Filter` objects
        :param indices: A list of indices indicating the corresponding filter in `filters`
        """
        if not isinstance(im_orig, Image):
            logger.warning(
                f"_apply_filters() passed {type(im_orig)} instead of Image instance"
            )
            # for now just convert it
            im_orig = Image(im_orig)

        im = im_orig.copy()

        for i, filt in enumerate(filters):
            idx_k = np.where(indices == i)[0]
            if len(idx_k) > 0:
                im[idx_k] = im[idx_k].filter(filt).asnumpy()

        return im

    def _apply_source_filters(self, im_orig, indices):
        return self._apply_filters(
            im_orig,
            self.unique_filters,
            self.filter_indices[indices],
        )

    @as_copy
    def cache(self):
        """
        Computes all queued pipeline transformations and stores the
        generated images in an array.  This trades memory for fast
        image access, and is useful when images will be repeatedly
        queried since it avoids recomputing on-the-fly.
        """
        logger.info("Caching source images")
        self._cached_im = self.images[:]
        self.generation_pipeline.reset()

    @property
    def images(self):
        """
        Subscriptable property which returns the images contained in this source
        corresponding to the indices given.
        """
        return self._img_accessor

    @abstractmethod
    def _images(self, indices):
        """
        Subclasses must implement a private _images() method accepting a 1-D NumPy array of indices.
        Subclasses handle cached image check as well as applying transforms in the generation pipeline.
        """

    @as_copy
    def downsample(self, L):
        if L > self.L:
            raise ValueError(
                "Max desired resolution {L} should be less than the current resolution {self.L}."
            )
        logger.info(f"Setting max. resolution of source = {L}")

        self.generation_pipeline.add_xform(Downsample(resolution=L))

        ds_factor = self.L / L
        self.unique_filters = [f.scale(ds_factor) for f in self.unique_filters]
        self.offsets /= ds_factor

        self.L = L

    @as_copy
    def whiten(self, noise_estimate=None):
        """
        Modify the `ImageSource` in-place by appending a whitening filter to the generation pipeline.

        When no `noise_estimate` is provided, a `WhiteNoiseEstimator`
        will be instantiated at this preprocessing stage behind the
        scenes.

        :param noise_estimate: Optional, `NoiseEstimator` or `Filter`. When
            passed a `NoiseEstimator` the `filter` attribute will be
            queried.  Alternatively, the noise PSD may be passed
            directly as a `Filter` object.
        :return: On return, the `ImageSource` object has been modified in place.
        """

        if noise_estimate is None:
            noise_filter = WhiteNoiseEstimator(self).filter
        elif isinstance(noise_estimate, NoiseEstimator):
            # Any NoiseEstimator instance should have a `filter`.
            noise_filter = noise_estimate.filter
        elif isinstance(noise_estimate, Filter):
            # We were given a `Filter` object.
            noise_filter = noise_estimate
        else:
            raise TypeError(
                f"Whiten passed {noise_estimate}"
                " instead of `NoiseEstimator` or `Filter`."
            )

        logger.info("Whitening source object")
        whiten_filter = PowerFilter(noise_filter, power=-0.5)

        logger.info("Transforming all CTF Filters into Multiplicative Filters")
        self.unique_filters = [
            MultiplicativeFilter(f, whiten_filter) for f in self.unique_filters
        ]
        logger.info("Adding Whitening Filter Xform to end of generation pipeline")
        self.generation_pipeline.add_xform(FilterXform(whiten_filter))

    @as_copy
    def phase_flip(self):
        """
        Perform phase flip on images in the source object using CTF information.
        If no CTFFilters exist this will emit a warning and otherwise no-op.
        """

        logger.info("Perform phase flip on source object")

        if len(self.unique_filters) >= 1:
            unique_xforms = [FilterXform(f.sign) for f in self.unique_filters]

            logger.info("Adding Phase Flip Xform to end of generation pipeline")
            self.generation_pipeline.add_xform(
                IndexedXform(unique_xforms, self.filter_indices)
            )

        else:
            # No CTF filters found
            logger.warning(
                "No Filters found."
                "  `phase_flip` is a no-op without Filters."
                "  Confirm you have correctly populated CTFFilters."
            )

    @as_copy
    def invert_contrast(self, batch_size=512):
        """
        invert the global contrast of images

        Check if all images in a stack should be globally phase flipped so that
        the molecule corresponds to brighter pixels and the background corresponds
        to darker pixels. This is done by comparing the mean in a small circle
        around the origin (supposed to correspond to the molecule) with the mean
        of the noise, and making sure that the mean of the molecule is larger.
        From the implementation level, we modify the `ImageSource` in-place by
        appending a `Multiple` filter to the generation pipeline.

        :param batch_size: Batch size of images to query.
        :return: On return, the `ImageSource` object has been modified in place.
        """

        logger.info("Apply contrast inversion on source object")
        L = self.L
        grid = grid_2d(L, indexing="yx", shifted=True)
        # Get mask indices of signal and noise samples assuming molecule
        signal_mask = grid["r"] < 0.5
        noise_mask = grid["r"] > 0.8

        # Calculate mean values in batch_size
        signal_mean = 0.0
        noise_mean = 0.0

        logger.info("Computing signal vs background contrast on source object")
        for i in trange(0, self.n, batch_size):
            images = self.images[i : i + batch_size].asnumpy()
            signal = images * signal_mask
            noise = images * noise_mask
            signal_mean += np.sum(signal)
            noise_mean += np.sum(noise)
        signal_denominator = self.n * np.sum(signal_mask)
        noise_denominator = self.n * np.sum(noise_mask)
        signal_mean /= signal_denominator
        noise_mean /= noise_denominator

        if signal_mean < noise_mean:
            logger.info("Need to invert contrast")
            scale_factor = -1.0
        else:
            logger.info("No need to invert contrast")
            scale_factor = 1.0

        logger.info("Adding Scaling Xform to end of generation pipeline")
        self.generation_pipeline.add_xform(Multiply(scale_factor))

    @as_copy
    def normalize_background(self, bg_radius=1.0, do_ramp=True):
        """
        Normalize the images by the noise background

        This is done by shifting the image density by the mean value of background
        and scaling the image density by the standard deviation of background.
        From the implementation level, we modify the `ImageSource` in-place by
        appending the `Add` and `Multiple` filters to the generation pipeline.

        :param bg_radius: Radius cutoff to be considered as background (in image size)
        :param do_ramp: When it is `True`, fit a ramping background to the data
            and subtract. Namely perform normalization based on values from each image.
            Otherwise, a constant background level from all images is used.
        :return: On return, the `ImageSource` object has been modified in place.
        """

        logger.info(
            f"Normalize background on source object with radius "
            f"size of {bg_radius} and do_ramp of {do_ramp}"
        )
        self.generation_pipeline.add_xform(
            LambdaXform(normalize_bg, bg_radius=bg_radius, do_ramp=do_ramp)
        )

    def im_backward(self, im, start, weights=None):
        """
        Apply adjoint mapping to set of images

        :param im: An Image instance to which we wish to apply the adjoint of the forward model.
        :param start: Start index of image to consider
        :param weights: Optional vector of weights to apply to images.
        Weights should be length `self.n`.
        :return: An L-by-L-by-L volume containing the sum of the adjoint mappings applied to the start+num-1 images.
        """
        num = im.n_images

        all_idx = np.arange(start, min(start + num, self.n))
        im *= self.amplitudes[all_idx, np.newaxis, np.newaxis]
        im = im.shift(-self.offsets[all_idx, :])
        im = self._apply_source_filters(im, all_idx)

        if weights is not None:
            im *= weights[all_idx, np.newaxis, np.newaxis]

        vol = im.backproject(self.rotations[start : start + num, :, :])[0]

        return vol

    def vol_forward(self, vol, start, num):
        """
        Apply forward image model to volume

        :param vol: A volume instance.
        :param start: Start index of image to consider
        :param num: Number of images to consider
        :return: The images obtained from volume by projecting, applying CTFs, translating, and multiplying by the
            amplitude.
        """
        all_idx = np.arange(start, min(start + num, self.n))
        if vol.n_vols != 1:
            raise ValueError("vol_forward expects a single volume, not a stack.")

        if vol.dtype != self.dtype:
            logger.warning(f"Volume.dtype {vol.dtype} inconsistent with {self.dtype}")

        im = vol.project(self.rotations[all_idx, :, :])
        im = self._apply_source_filters(im, all_idx)
        im = im.shift(self.offsets[all_idx, :])
        im *= self.amplitudes[all_idx, np.newaxis, np.newaxis]
        return im

    def save(
        self,
        starfile_filepath,
        batch_size=512,
        save_mode=None,
        overwrite=False,
    ):
        """
        Save the output metadata to STAR file and/or images to MRCS file.

        :param starfile_filepath: Path to STAR file where we want to
            save metadata of image_source
        :param batch_size: Batch size of images to query.
            Note, `batch_size=1` implies single MRC extension `.mrc`,
            while `batch_size>=1` implies stack MRC extension `.mrcs`.
        :param save_mode: Whether to save all images in a `single` or multiple files in batch size.
            Default is multiple, supply `'single'` for single mode.
        :param overwrite: Option to overwrite the output MRC files.
        :return: A dictionary containing "starfile"--the path to the saved starfile-- and "mrcs", a
            list of the saved particle stack MRC filenames.
        """
        logger.info("save metadata into STAR file")
        filename_indices = self.save_metadata(
            starfile_filepath,
            batch_size=batch_size,
            save_mode=save_mode,
        )
        unique_filenames = list(dict.fromkeys(filename_indices))

        logger.info("save images into MRCS file")
        self.save_images(
            starfile_filepath,
            filename_indices=filename_indices,
            batch_size=batch_size,
            overwrite=overwrite,
        )
        # return some information about the saved files
        info = {"starfile": starfile_filepath, "mrcs": unique_filenames}
        return info

    @staticmethod
    def _populate_common_metadata(
        n,
        meta_dict,
        local_cols,
        starfile_filepath,
        batch_size,
        save_mode,
    ):
        """
        Populate metadata columns common to all `ImageSource` subclasses.
        """
        # Create a new key that we will be populating in the loop below
        meta_dict["_rlnImageName"] = np.full(n, fill_value="").astype("object")

        if save_mode == "single":
            # Save all images into one single mrc file
            fname = os.path.basename(starfile_filepath)
            fstem = os.path.splitext(fname)[0]
            mrcs_filename = f"{fstem}_{0}_{n-1}.mrcs"

            # Then set name in dict for the StarFile
            meta_dict["_rlnImageName"][:] = [
                f"{j + 1:06}@{mrcs_filename}" for j in range(n)
            ]
        else:
            # save all images into multiple mrc files in batch size
            # When batch_size is >1, use the plural extension.
            ext = ".mrcs"
            if batch_size == 1:
                ext = ".mrc"

            for i_start in np.arange(0, n, batch_size):
                i_end = min(n, i_start + batch_size)
                num = i_end - i_start
                mrcs_filename = (
                    os.path.splitext(os.path.basename(starfile_filepath))[0]
                    + f"_{i_start}_{i_end-1}{ext}"
                )
                meta_dict["_rlnImageName"][i_start:i_end] = [
                    "{0:06}@{1}".format(j + 1, mrcs_filename) for j in range(num)
                ]

        # Subclass-specific columns are popped to the end of the dictionary in order:
        # pop() both removes the given column and returns its data as a ndarray,
        # which is then tacked back on to the rightmost side of metadata
        # Note that all dictionaries in py>=3.7 are ordered
        for col in local_cols:
            meta_dict[col] = meta_dict.pop(col)

    def _populate_local_metadata(self):
        """
        Populate metadata columns specific to the `ImageSource` subclass being saved.
        Subclasses optionally override, but must return a list of strings.
        :return: A list of the names of the columns added.
        """
        return []

    def save_metadata(self, starfile_filepath, batch_size=512, save_mode=None):
        """
        Save updated metadata to a STAR file

        :param starfile_filepath: Path to STAR file where we want to
            save image_source
        :param batch_size: Batch size of images to query from the
            `ImageSource` object. Every `batch_size` rows, entries are
            written to STAR file.
        :param save_mode: Whether to save all images in a single or
            multiple files in batch size.
        :return: None
        """

        # Get local metadata columns that were added by subclass
        local_cols = self._populate_local_metadata()

        metadata = self.get_metadata(as_dict=True).copy()
        # Drop any column that doesn't start with a *single* underscore
        metadata = {
            k: v
            for k, v in metadata.items()
            if k.startswith("_") and not k.startswith("__")
        }

        # Populates _rlnImageName column, setting up filepaths to .mrcs stacks
        self._populate_common_metadata(
            self.n, metadata, local_cols, starfile_filepath, batch_size, save_mode
        )

        filename_indices = [
            x[1]
            for x in np.char.split(metadata["_rlnImageName"].astype(np.str_), sep="@")
        ]

        # initialize the star file object and save it
        odict = OrderedDict()
        # since our StarFile only has one block, the convention is to save it with the header "data_", i.e. its name is blank
        # if we had a block called "XYZ" it would be saved as "XYZ"
        # thus we index the metadata block with ""
        odict[""] = metadata
        out_star = StarFile(blocks=odict)
        out_star.write(starfile_filepath)
        return filename_indices

    def save_images(
        self, starfile_filepath, filename_indices=None, batch_size=512, overwrite=False
    ):
        """
        Save an ImageSource to MRCS files

        Note that .mrcs files are saved at the same location as the STAR file.

        :param filename_indices: Filename list for save all images
        :param starfile_filepath: Path to STAR file where we want to save image_source
        :param batch_size: Batch size of images to query from the `ImageSource` object.
            if `save_mode` is not `single`, images in the same batch will save to one MRCS file.
        :param overwrite: Whether to overwrite any .mrcs files found at the target location.
        :return: None
        """

        if filename_indices is None:
            # Generate filenames from metadata
            filename_indices = [
                self._metadata["_rlnImageName"][i].split("@")[1] for i in range(self.n)
            ]

        # get the save_mode from the file names
        unique_filename = set(filename_indices)
        save_mode = None
        if len(unique_filename) == 1:
            save_mode = "single"

        if save_mode == "single":
            # Save all images into one single mrc file

            # First, construct name for mrc file
            fdir = os.path.dirname(starfile_filepath)
            mrcs_filepath = os.path.join(fdir, filename_indices[0])

            # Open new MRC file
            with mrcfile.new_mmap(
                mrcs_filepath,
                shape=(self.n, self.L, self.L),
                mrc_mode=2,
                overwrite=overwrite,
            ) as mrc:
                stats = MrcStats()
                # Loop over source setting data into mrc file
                for i_start in np.arange(0, self.n, batch_size):
                    i_end = min(self.n, i_start + batch_size)
                    logger.info(
                        f"Saving ImageSource[{i_start}-{i_end-1}] to {mrcs_filepath}"
                    )
                    datum = self.images[i_start:i_end].asnumpy().astype("float32")

                    # Assign to mrcfile
                    mrc.data[i_start:i_end] = datum

                    # Accumulate stats
                    stats.push(datum)

                # To be safe, explicitly set the header
                #   before the mrc file context closes.
                mrc.update_header_from_data()

                # Also write out updated statistics for this mrc.
                #   This should be an optimization over mrc.update_header_stats
                #   for large arrays.
                stats.update_header(mrc)

        else:
            # save all images into multiple mrc files in batch size
            for i_start in np.arange(0, self.n, batch_size):
                i_end = min(self.n, i_start + batch_size)

                mrcs_filepath = os.path.join(
                    os.path.dirname(starfile_filepath), filename_indices[i_start]
                )

                logger.info(
                    f"Saving ImageSource[{i_start}-{i_end-1}] to {mrcs_filepath}"
                )
                im = self.images[i_start:i_end]
                im.save(mrcs_filepath, overwrite=overwrite)

    def estimate_signal_mean_energy(
        self,
        sample_n=None,
        support_radius=None,
        batch_size=512,
        image_accessor=None,
    ):
        """
        Estimate the signal mean of `sample_n` projections.

        :param sample_n: Number of images used for estimate.
            Defaults to all images in source.
        :param support_radius: Pixel radius used for masking signal support.
            Default of None will compute inscribed circle, `self.L // 2`.
        :param batch_size: Images per batch, defaults 512.
        :param image_accessor: Optionally override images. Defaults `self.images`.
        :returns: Estimated signal mean
        """

        if sample_n is None:
            sample_n = self.n

        if sample_n > self.n:
            logger.warning(
                f"`estimate_signal_mean_energy` sample_n > Source.n: {sample_n} > {self.n}."
                f" Accuracy may be impaired, setting sample_n=self.n={self.n}"
            )
            sample_n = self.n

        images = image_accessor or self.images

        mask = support_mask(self.L, support_radius, dtype=self.dtype)

        # mean is estimated batch-wise, compare with numpy
        s = 0.0
        _denom = sample_n * np.sum(mask)
        for i in trange(0, sample_n, batch_size):
            # Gather this batch of images and mask off area outside support_radius
            images_masked = images[i : i + batch_size].asnumpy()[..., mask]
            # Accumulate second moments
            s += np.sum(images_masked**2) / _denom

        logger.debug(f"Source estimated signal mean energy: {s}")

        return s

    def estimate_signal_var(
        self, sample_n=None, support_radius=None, batch_size=512, image_accessor=None
    ):
        """
        Estimate the signal variance of `sample_n` projections.

        :param sample_n: Number of images used for estimate.
            Defaults to all images in source.
        :param support_radius: Pixel radius used for masking signal support.
            Default of None will compute inscribed circle, `self.L // 2`.
        :param batch_size: Images per batch, defaults 512.
        :param image_accessor: Optionally override images. Defaults `self.images`.
        :returns: Estimated signal variance.
        """

        if sample_n is None:
            sample_n = self.n

        if sample_n > self.n:
            logger.warning(
                f"`estimate_signal_var` sample_n > Source.n: {sample_n} > {self.n}."
                f" Accuracy may be impaired, setting sample_n=self.n={self.n}"
            )
            sample_n = self.n

        images = image_accessor or self.images

        mask = support_mask(self.L, support_radius, dtype=self.dtype)

        # Var is estimated batch-wise, compare with numpy
        # np_estimated_var = np.var(images[:sample_n].asnumpy()[..., mask])
        first_moment = 0.0
        second_moment = 0.0
        _denom = sample_n * np.sum(mask)
        for i in trange(0, sample_n, batch_size):
            # Gather this batch of images and mask off area outside support_radius
            images_masked = images[i : i + batch_size].asnumpy()[..., mask]
            # Accumulate first and second moments
            first_moment += np.sum(images_masked) / _denom
            second_moment += np.sum(images_masked**2) / _denom

        # E[X**2] - E[X]**2
        estimated_var = second_moment - first_moment**2
        logger.debug(f"Source estimated signal var: {estimated_var}")

        return estimated_var

    def estimate_signal_power(
        self,
        sample_n=None,
        support_radius=None,
        batch_size=512,
        signal_power_method="estimate_signal_mean_energy",
        image_accessor=None,
    ):
        """
        Estimate the signal energy of `sample_n` projections using prescribed method.

        :param sample_n: Number of images used for estimate.
            Defaults to all images in source.
        :param support_radius: Pixel radius used for masking signal support.
            Default of None will compute inscribed circle, `self.L // 2`.
        :param batch_size: Images per batch, defaults 512.
        :param signal_power_method: Method used for computing signal energy.
           Defaults to mean via `estimate_signal_mean_energy`.
           Can use variance method via `estimate_signal_var`.
        :param image_accessor: Optionally override images. Defaults `self.images`.
        :returns: Estimated signal variance.
        """

        try:
            signal_estimate_method = getattr(self, signal_power_method)
        except AttributeError:
            raise ValueError(
                f"Cannot find signal_power_method={signal_power_method}."
                "  Try the default 'estimate_signal_mean_energy' or 'estimate_signal_var'"
            )

        signal_power = signal_estimate_method(
            sample_n=sample_n,
            support_radius=support_radius,
            batch_size=batch_size,
            image_accessor=image_accessor,
        )

        return signal_power

    def estimate_noise_power(
        self,
        sample_n=None,
        support_radius=None,
        batch_size=512,
    ):
        """
        Estimate the noise energy of `sample_n` images using
        `WhiteNoiseEstimator`.

        :param sample_n: Number of images used for estimate.
            Defaults to all images in source.
        :param support_radius: Pixel radius used for masking signal support.
            Default of None will compute inscribed circle, `self.L // 2`.
        :param batch_size: Images per batch, defaults 512.

        :returns: Estimated noise energy (variance).
        """

        if support_radius is None:
            support_radius_proportion = 1
        else:
            # Note, noise_estimator expects radius as proportion.
            support_radius_proportion = support_radius / (self.L // 2)

        est = WhiteNoiseEstimator(
            src=self, bgRadius=support_radius_proportion, batchSize=batch_size
        )

        return est.estimate()

    def estimate_snr(
        self,
        sample_n=None,
        support_radius=None,
        batch_size=512,
        noise_power=None,
        signal_power_method="estimate_signal_mean_energy",
    ):
        """
        Estimate the SNR of the simulated data set using
        estimated signal power / noise power.

        Note signal power depends on choice of `signal_power_method`,
        but differences should be small in practice when background
        noise is zero centered.

        :param sample_n: Number of images used for estimate.
            Defaults to all images in source.
        :param support_radius: Pixel radius used for masking signal support.
            Default of None will compute inscribed circle, `self.L // 2`.
        :param batch_size: Images per batch, defaults 512.
        :param signal_power_method: Method used for computing signal energy.
           Defaults to mean via `estimate_signal_mean_energy`.
           Can use variance method via `estimate_signal_var`.
        :returns: Estimated signal to noise ratio.
        """

        if sample_n is None:
            sample_n = self.n

        if sample_n > self.n:
            logger.warning(
                f"`estimate_snr` sample_n > Source.n: {sample_n} > {self.n}."
                f" Accuracy may be impaired, setting sample_n=self.n={self.n}"
            )
            sample_n = self.n

        if noise_power is None:
            noise_power = self.estimate_noise_power()

        signal_power = self.estimate_signal_power(
            sample_n=sample_n,
            support_radius=support_radius,
            batch_size=batch_size,
            signal_power_method=signal_power_method,
        )

        # For `estimate_signal_mean_energy` we yield: mean(signal**2)
        #     `estimate_signal_var`   we yield: signal_variance
        snr = (signal_power - noise_power) / noise_power

        # Check for extremal values.
        if snr < 0:
            logger.warning(
                "For extremely low SNR, estimation accuracy may be impaired."
                f"  Clamping estimated SNR {snr} to 0."
            )
            snr = 0

        return snr


class IndexedSource(ImageSource):
    """
    Map into another into ImageSource.
    """

    def __init__(self, src, indices, memory=None):
        """
        Instantiates a new source along given `indices`.

        :param src: ImageSource to be used as the source.
        :param index_map: index_map
        :param memory: str or None
            The path of the base directory to use as a data store or
            None. If None is given, no caching is performed.
        """

        self.src = src
        if not isinstance(src, ImageSource):
            raise TypeError(f"Input src {src} must be an ImageSource.")

        # `_ImageAccessor` performs checking and slicing logic.
        # `index_map` sequence forms a natural map from the "new" source -> "self".
        # Example, if request=slice(500,1000),
        #   then new_src[0] ~> old_src[500]; index_map[0] = 500.
        self.index_map = _ImageAccessor(lambda x: x, src.n)[indices]

        # Get all the metadata associated with these indices.
        metadata = self.src.get_metadata(indices=self.index_map, as_dict=True).copy()

        # Construct a fully formed ImageSource with this metadata
        super().__init__(
            L=src.L,
            n=len(self.index_map),
            dtype=src.dtype,
            metadata=metadata,
            memory=memory,
        )

        # Create filter indices, these are required to pass unharmed through filter eval code
        #   that is potentially called by other methods later.
        self.filter_indices = np.zeros(self.n, dtype=int)
        self.unique_filters = [IdentityFilter()]

        # Any further operations should not mutate this instance.
        self._mutable = False

    def _images(self, indices):
        """
        Returns images from `self.src` corresponding to `indices`
        remapped by `self.index_map`.

        :param indices: A 1-D NumPy array of indices.
        :return: An `Image` object.
        """
        mapped_indices = self.index_map[indices]
        # Load previous source image data and apply any transforms
        # belonging to this IndexedSource.  Note the previous source
        # requires remapped indices, while the current source uses the
        # `indices` arg directly.
        return self.generation_pipeline.forward(
            self.src.images[mapped_indices], indices
        )

    def __repr__(self):
        return f"{self.__class__.__name__} mapping {self.n} of {self.src.n} indices from {self.src.__class__.__name__}."


class OrientedSource(IndexedSource):
    """
    Source for oriented 2D images using orientation estimation methods.
    """

    def __init__(self, src, orientation_estimator=None):
        """
        Constructor of an oriented ImageSource object. Orientation is determined by
        performing orientation estimation using a supplied `orientation_estimator`.

        :param src: Source used for orientation estimation
        :param orientation_estimator: CLOrient3D subclass used for orientation estimation.
            Default uses the CLSyncVoting method.
        """

        self.src = src
        if not isinstance(self.src, ImageSource):
            raise ValueError(
                f"`src` should be subclass of `ImageSource`, found {self.src}."
            )

        # `indices` for IndexedSource.
        indices = np.arange(self.src.n)

        super().__init__(
            src=self.src,
            indices=indices,
        )

        # Remove any orientation information passed in by original source.
        self._reset_orientation()

        # Flag for lazy eval of orientation estimation.
        self._oriented = False

        if orientation_estimator is None:
            orientation_estimator = CLSyncVoting(src)

        self.orientation_estimator = orientation_estimator
        if not isinstance(self.orientation_estimator, CLOrient3D):
            raise ValueError(
                "`orientation_estimator` should be subclass of `CLOrient3D`,"
                f" found {self.orientation_estimator}."
            )

        # Any further operations should not mutate this instance.
        self._mutable = False

    def _orient(self):
        """
        Perform orientation estimation if not already done.
        """

        # Short circuit.
        if self._oriented:
            logger.debug(f"{self.__class__.__name__} already oriented, skipping")
            return

        logger.info(
            f"Estimating rotations for {self.src} using {self.orientation_estimator}."
        )
        self.orientation_estimator.estimate_rotations()

        # Allow mutability to set rotations.
        self._mutable = True
        self.rotations = self.orientation_estimator.rotations
        self._mutable = False

        self._oriented = True

    def _reset_orientation(self):
        """
        Remove orientation information passed in by original source.
        """
        _info_removed = False
        rot_keys = ["_rlnAngleRot", "_rlnAngleTilt", "_rlnAnglePsi"]
        for key in rot_keys:
            if self.has_metadata(key):
                del self._metadata[key]
                _info_removed = True
        if _info_removed:
            logger.info(f"Removing orientation information passed by {self.src}.")

    def _rots(self):
        """
        Converts internal `_rotations` representation to expected matrix form.
        If rotations have not been set orientation estimation will be performed
        prior to serving up the rotations.

        :return: Rotation matrices as a n x 3 x 3 array
        """

        self._orient()
        return self._rotations.matrices.astype(self.dtype)

    def _angles(self):
        self._orient()
        return super()._angles()

    def save_metadata(self, starfile_filepath, batch_size=512, save_mode=None):
        self._orient()
        return super().save_metadata(
            starfile_filepath, batch_size=batch_size, save_mode=None
        )

    def get_metadata(
        self, metadata_fields=None, indices=None, default_value=None, as_dict=False
    ):
        # get_metadata is used during ImageSource instantiation, so only perform lazy
        # orientation estimation if the oriented_source is already initialized,
        # ie. when no longer mutable.
        if not self._mutable:
            self._orient()
        return super().get_metadata(
            metadata_fields=metadata_fields,
            indices=indices,
            default_value=default_value,
            as_dict=as_dict,
        )

    def __repr__(self):
        return f"{self.__class__.__name__} for origin source {self.src.__class__.__name__}."


class ArrayImageSource(ImageSource):
    """
    An `ImageSource` object that holds a reference to an underlying `Image` object (a thin wrapper on an ndarray)
    representing images. It does not produce its images on the fly, but keeps them in memory. As such, it should not be
    used where large Image objects are involved, but can be used in situations where API conformity is desired.

    Note that this class does not provide an `_images` method, since it populates the `_cached_im` attribute which,
    if available, is consulted directly by the parent class, bypassing `_images`.
    """

    def __init__(self, im, metadata=None, angles=None, symmetry_group=None):
        """
        Initialize from an `Image` object.

        :param im: An `Image` or Numpy array object representing image data served up by this `ImageSource`.
            In the case of a Numpy array, attempts to create an 'Image' object.
        :param metadata: A Dataframe of metadata information corresponding to this ImageSource's images
        :param angles: Optional n-by-3 array of rotation angles corresponding to `im`.
        :param symmetry_group: A SymmetryGroup instance or string indicating the underlying symmetry of the molecule.
        """

        if not isinstance(im, Image):
            logger.info("Attempting to create an Image object from Numpy array.")
            try:
                im = Image(im)
            except Exception as e:
                raise RuntimeError(
                    "Creating Image object from Numpy array failed."
                    f" Original error: {str(e)}"
                )

        super().__init__(
            L=im.resolution,
            n=im.n_images,
            dtype=im.dtype,
            metadata=metadata,
            memory=None,
            symmetry_group=symmetry_group,
        )

        self._cached_im = im

        # Create filter indices, these are required to pass unharmed through filter eval code
        #   that is potentially called by other methods later.
        self.filter_indices = np.zeros(self.n, dtype=int)
        self.unique_filters = [IdentityFilter()]

        # Optionally populate angles/rotations.
        if angles is not None:
            if angles.shape != (self.n, 3):
                raise ValueError(f"Angles should be shape {(self.n, 3)}")
            # This will populate `_rotations`,
            #   which is exposed by properties `angles` and `rotations`.
            self.angles = angles

        # Any further operations should not mutate this instance.
        self._mutable = False

    def _images(self, indices):
        """
        Returns images corresponding to `indices` after being accessed via the
        `ImageSource.images` property
        :param indices: A 1-D NumPy array of indices.
        :return: An `Image` object.
        """
        # Load cached data and apply transforms
        return self.generation_pipeline.forward(self._cached_im[indices, :, :], indices)

    def _rots(self):
        """
        Private method, checks if `_rotations` has been set,
        then returns inherited rotations, otherwise raise.
        """

        if self._rotations is not None:
            return super()._rots()
        else:
            raise RuntimeError(
                "Consumer of ArrayImageSource trying to access rotations,"
                " but rotations were not defined for this source."
                "  Try instantiating with angles."
            )

    def _angles(self):
        """
        Private method, checks if `_rotations` has been set,
        then returns inherited angles, otherwise raise.
        """

        if self._rotations is not None:
            return super()._angles()
        else:
            raise RuntimeError(
                "Consumer of ArrayImageSource trying to access angles,"
                " but angles were not defined for this source."
                "  Try instantiating with angles."
            )



import logging

from aspire import config

from .complex_pca.complex_pca import ComplexPCA

logger = logging.getLogger(__name__)


def numeric_object(which):
    if which == "cupy":
        from .cupy import Cupy as NumericClass
    elif which == "numpy":
        from .numpy import Numpy as NumericClass
    else:
        raise RuntimeError(f"Invalid selection for numeric module: {which}")
    return NumericClass()


xp = numeric_object(config["common"]["numeric"].as_str())


def fft_object(which):
    if which == "pyfftw":
        from .pyfftw_fft import PyfftwFFT as FFTClass
    elif which == "cupy":
        from .cupy_fft import CupyFFT as FFTClass
    elif which == "scipy":
        from .scipy_fft import ScipyFFT as FFTClass
    elif which == "mkl":
        from .mkl_fft import MKLFFT as FFTClass
    else:
        raise RuntimeError(f"Invalid selection for fft class: {which}")
    return FFTClass()


fft = fft_object(config["common"]["fft"].as_str())



import logging

import numpy as np
import scipy.sparse as sps
from scipy.linalg import qr

from aspire.utils import random

logger = logging.getLogger(__name__)


def pca_y(x, k, num_iters=2, seed=None):
    """
    PCA using QR factorization.

    See:

    An algorithm for the principal component analysis of large data sets.
    Halko, Martinsson, Shkolnisky, Tygert , SIAM 2011.

    :param x: Data matrix
    :param k: Number of estimated Principal Components.
    :param num_iters: Number of dot product applications.
    :return: (left Singular Vectors, Singular Values, right Singular Vectors)
    """

    m, n = x.shape

    def operator(mat):
        return x.dot(mat)

    def operator_transpose(mat):
        return np.conj(x.T).dot(mat)

    flag = False
    if m < n:
        flag = True
        operator_transpose, operator = operator, operator_transpose
        m, n = n, m

    ones = np.ones((n, k + 2))
    if x.dtype == np.dtype("complex"):
        h = operator(
            (2 * random((k + 2, n), seed=seed).T - ones)
            + 1j * (2 * random((k + 2, n), seed=seed).T - ones)
        )
    else:
        h = operator(2 * random((k + 2, n), seed=seed).T - ones)

    f = [h]

    for _ in range(num_iters):
        h = operator_transpose(h)
        h = operator(h)
        f.append(h)

    f = np.concatenate(f, axis=1)
    # f has e-16 error, q has e-13
    q, _, _ = qr(f, mode="economic", pivoting=True)
    b = np.conj(operator_transpose(q)).T
    u, s, v = np.linalg.svd(b, full_matrices=False)
    # not sure how to fix the signs but it seems like I dont need to
    # TODO use fix_svd, here and matlab
    # u, v = fix_svd(u, v)

    v = v.conj()
    u = np.dot(q, u)

    u = u[:, :k]
    v = v[:k]
    s = s[:k]

    if flag:
        u, v = v.T, u.T

    return u, s, v


def bispec_operator_1(freqs):
    max_freq = np.max(freqs)
    count = 0
    for i in range(2, max_freq):
        for j in range(1, min(i, max_freq - i + 1)):
            k = i + j
            id1 = np.where(freqs == i)[0]
            id2 = np.where(freqs == j)[0]
            id3 = np.where(freqs == k)[0]
            nd1 = len(id1)
            nd2 = len(id2)
            nd3 = len(id3)
            count += nd1 * nd2 * nd3

    full_list = np.zeros((count, 3), dtype="int")
    count = 0
    for i in range(2, max_freq):
        for j in range(1, min(i, max_freq - i + 1)):
            k = i + j
            id1 = np.where(freqs == i)[0]
            id2 = np.where(freqs == j)[0]
            id3 = np.where(freqs == k)[0]
            nd1 = len(id1)
            nd2 = len(id2)
            nd3 = len(id3)
            nd = nd1 * nd2 * nd3
            if nd != 0:
                tmp1 = np.tile(id1, nd2)
                tmp2 = np.repeat(id2, nd1)
                tmp = np.stack((tmp1, tmp2), axis=1)
                tmp1 = np.tile(tmp, (nd3, 1))
                tmp2 = np.repeat(id3, nd1 * nd2)
                full_list[count : count + nd, :2] = tmp1
                full_list[count : count + nd, 2] = tmp2
                count += nd

    val = np.ones(full_list.shape)
    val[:, 2] = -1  # conjugation
    n_col = count
    full_list = full_list.flatten("F")
    val = val.flatten("F")
    col = np.tile(np.arange(n_col), 3)
    o1 = sps.csr_matrix(
        (np.ones(len(full_list)), (col, full_list)), shape=(n_col, len(freqs))
    )
    o2 = sps.csr_matrix((val, (col, full_list)), shape=(n_col, len(freqs)))
    return o1, o2


def bispec_2drot_large(coeff, freqs, eigval, alpha, sample_n, seed=None):
    """
    alpha 1/3
    sample_n 4000
    """
    freqs_not_zero = freqs != 0

    coeff_norm = np.log(np.power(np.absolute(coeff[freqs_not_zero]), alpha))
    if np.any(coeff_norm == float("-inf")):
        raise ValueError("coeff_norm should not be -inf")

    phase = coeff[freqs_not_zero] / np.absolute(coeff[freqs_not_zero])
    phase = np.arctan2(np.imag(phase), np.real(phase))
    eigval = eigval[freqs_not_zero]
    o1, o2 = bispec_operator_1(freqs[freqs_not_zero])

    # GBW, naively handle vanishing eigvals.
    #  This became a problem with very noisy images...
    p = np.power(eigval, alpha)
    mask = np.where(p, p, -1)  # taking the log in the next step will yield a 0
    m = np.exp(o1 * np.log(p, where=(mask > 0)))
    p_m = m / m.sum()
    x = random(size=len(m), seed=seed)
    m_id = np.where(x < sample_n * p_m)[0]
    o1 = o1[m_id]
    o2 = o2[m_id]
    m = np.exp(o1 * coeff_norm + 1j * o2 * phase)

    # svd of the reduced bispectrum
    u, s, v = pca_y(m, 300, seed=seed)

    coeff_b = np.einsum("i, ij -> ij", s, np.conjugate(v))
    coeff_b_r = np.conjugate(u.T).dot(np.conjugate(m))

    coeff_b = coeff_b / np.linalg.norm(coeff_b, axis=0)
    coeff_b_r = coeff_b_r / np.linalg.norm(coeff_b_r, axis=0)

    return coeff_b, coeff_b_r



import logging

import numpy as np
from scipy.special import jv

from aspire.basis import FBBasisMixin, SteerableBasis2D
from aspire.basis.basis_utils import unique_coords_nd
from aspire.utils import complex_type, real_type, roll_dim, unroll_dim
from aspire.utils.matlab_compat import m_flatten, m_reshape

logger = logging.getLogger(__name__)


class FBBasis2D(SteerableBasis2D, FBBasisMixin):
    """
    Define a derived class using the Fourier-Bessel basis for mapping 2D images

    The expansion coefficients of 2D images on this basis are obtained by
    the least squares method. The algorithm is described in the publication:
    Z. Zhao, A. Singer, Fourier-Bessel Rotational Invariant Eigenimages,
    The Journal of the Optical Society of America A, 30 (5), pp. 871-877 (2013).

    """

    # TODO: Methods that return dictionaries should return useful objects instead
    def __init__(self, size, ell_max=None, dtype=np.float32):
        """
        Initialize an object for the 2D Fourier-Bessel basis class

        :param size: The size of the vectors for which to define the basis.
            May be a 2-tuple or an integer, in which case a square basis is assumed.
            Currently only square images are supported.
        :ell_max: The maximum order ell of the basis elements. If no input
            (= None), it will be set to np.Inf and the basis includes all
            ell such that the resulting basis vectors are concentrated
            below the Nyquist frequency (default Inf).
        """

        if isinstance(size, int):
            size = (size, size)
        ndim = len(size)
        assert ndim == 2, "Only two-dimensional basis functions are supported."
        assert len(set(size)) == 1, "Only square domains are supported."
        super().__init__(size, ell_max, dtype=dtype)

    def _build(self):
        """
        Build the internal data structure to 2D Fourier-Bessel basis
        """
        logger.info(
            "Expanding 2D images in a spatial-domain Fourier–Bessel"
            " basis using the direct method."
        )

        # get upper bound of zeros, ells, and ks  of Bessel functions
        self._calc_k_max()

        # calculate total number of basis functions
        self.count = self.k_max[0] + sum(2 * self.k_max[1:])

        # obtain a 2D grid to represent basis functions
        self.basis_coords = unique_coords_nd(self.nres, self.ndim, dtype=self.dtype)

        # generate 1D indices for basis functions
        self._compute_indices()
        self._indices = self.indices()

        # get normalized factors
        self.radial_norms, self.angular_norms = self.norms()

        # precompute the basis functions in 2D grids
        self._precomp = self._precomp()

    def _compute_indices(self):
        """
        Create the indices for each basis function
        """
        indices_ells = np.zeros(self.count, dtype=int)
        indices_ks = np.zeros(self.count, dtype=int)
        indices_sgns = np.zeros(self.count, dtype=int)

        # We'll also generate a mapping for complex construction
        self.complex_count = sum(self.k_max)
        # These map indices in complex array to pair of indices in real array
        self._pos = np.zeros(self.complex_count, dtype=int)
        self._neg = np.zeros(self.complex_count, dtype=int)

        i = 0
        ci = 0
        for ell in range(self.ell_max + 1):
            sgns = (1,) if ell == 0 else (1, -1)
            ks = np.arange(0, self.k_max[ell])

            for sgn in sgns:
                rng = np.arange(i, i + len(ks))
                indices_ells[rng] = ell
                indices_ks[rng] = ks
                indices_sgns[rng] = sgn

                if sgn == 1:
                    self._pos[ci + ks] = rng
                elif sgn == -1:
                    self._neg[ci + ks] = rng

                i += len(ks)

            ci += len(ks)

        self.angular_indices = indices_ells
        self.radial_indices = indices_ks
        self.signs_indices = indices_sgns
        # Relating to paper: a[i] = a_ell_ks = a_angularindices[i]_radialindices[i]
        self.complex_angular_indices = indices_ells[self._pos]  # k
        self.complex_radial_indices = indices_ks[self._pos]  # q

    def indices(self):
        """
        Return the precomputed indices for each basis function.
        """
        return {
            "ells": self.angular_indices,
            "ks": self.radial_indices,
            "sgns": self.signs_indices,
        }

    def _precomp(self):
        """
        Precompute the basis functions at defined sample points
        """

        r_unique = self.basis_coords["r_unique"]
        ang_unique = self.basis_coords["ang_unique"]

        ind_radial = 0
        ind_ang = 0

        radial = np.zeros(shape=(len(r_unique), np.sum(self.k_max)), dtype=self.dtype)
        ang = np.zeros(
            shape=(ang_unique.shape[-1], 2 * self.ell_max + 1), dtype=self.dtype
        )

        for ell in range(0, self.ell_max + 1):
            for k in range(1, self.k_max[ell] + 1):
                # Only normalized by the radial part of basis function
                radial[:, ind_radial] = (
                    jv(ell, self.r0[ell][k - 1] * r_unique)
                    / self.radial_norms[ind_radial]
                )
                ind_radial += 1

            sgns = (1,) if ell == 0 else (1, -1)
            for sgn in sgns:
                fn = np.cos if sgn == 1 else np.sin
                ang[:, ind_ang] = fn(ell * ang_unique)
                ind_ang += 1

        return {"radial": radial, "ang": ang}

    def norms(self):
        """
        Calculate the normalized factors of basis functions
        """
        radial_norms = np.zeros(np.sum(self.k_max), dtype=self.dtype)
        angular_norms = np.zeros(np.sum(self.k_max), dtype=self.dtype)
        norm_fn = self.basis_norm_2d

        i = 0
        for ell in range(0, self.ell_max + 1):
            for k in range(1, self.k_max[ell] + 1):
                radial_norms[i], angular_norms[i] = norm_fn(ell, k)
                i += 1

        return radial_norms, angular_norms

    def basis_norm_2d(self, ell, k):
        """
        Calculate the normalized factors from radial and angular parts of a specified basis function
        """
        rad_norm = (
            np.abs(jv(ell + 1, self.r0[ell][k - 1]))
            * np.sqrt(1 / 2.0)
            * self.nres
            / 2.0
        )
        ang_norm = np.sqrt(np.pi)
        if ell == 0:
            ang_norm *= np.sqrt(2)

        return rad_norm, ang_norm

    def _evaluate(self, v):
        """
        Evaluate coefficients in standard 2D coordinate basis from those in FB basis

        :param v: A coefficient vector (or an array of coefficient vectors) to
            be evaluated. The last dimension must equal `self.count`.
        :return: The evaluation of the coefficient vector(s) `v` for this basis.
            This is an array whose last dimensions equal `self.sz` and the remaining
            dimensions correspond to first dimensions of `v`.
        """
        # Transpose here once, instead of several times below  #RCOPT
        v = v.reshape(-1, self.count).T

        r_idx = self.basis_coords["r_idx"]
        ang_idx = self.basis_coords["ang_idx"]
        mask = m_flatten(self.basis_coords["mask"])

        ind = 0
        ind_radial = 0
        ind_ang = 0

        x = np.zeros(shape=tuple([np.prod(self.sz)] + list(v.shape[1:])), dtype=v.dtype)
        for ell in range(0, self.ell_max + 1):
            k_max = self.k_max[ell]
            idx_radial = ind_radial + np.arange(0, k_max, dtype=int)

            # include the normalization factor of angular part
            ang_nrms = self.angular_norms[idx_radial]
            radial = self._precomp["radial"][:, idx_radial]
            radial = radial / ang_nrms

            sgns = (1,) if ell == 0 else (1, -1)
            for _ in sgns:
                ang = self._precomp["ang"][:, ind_ang]
                ang_radial = np.expand_dims(ang[ang_idx], axis=1) * radial[r_idx]
                idx = ind + np.arange(0, k_max, dtype=int)
                x[mask] += ang_radial @ v[idx]
                ind += len(idx)
                ind_ang += 1

            ind_radial += len(idx_radial)

        x = x.T.reshape(-1, *self.sz)  # RCOPT

        return x

    def _evaluate_t(self, v):
        """
        Evaluate coefficient in FB basis from those in standard 2D coordinate basis

        :param v: The coefficient array to be evaluated. The last dimensions
            must equal `self.sz`.
        :return: The evaluation of the coefficient array `v` in the dual basis
            of `basis`. This is an array of vectors whose last dimension equals
            `self.count` and whose first dimensions correspond to
            first dimensions of `v`.
        """
        v = v.T
        x, sz_roll = unroll_dim(v, self.ndim + 1)
        x = m_reshape(
            x, new_shape=tuple([np.prod(self.sz)] + list(x.shape[self.ndim :]))
        )

        r_idx = self.basis_coords["r_idx"]
        ang_idx = self.basis_coords["ang_idx"]
        mask = m_flatten(self.basis_coords["mask"])

        ind = 0
        ind_radial = 0
        ind_ang = 0

        v = np.zeros(shape=tuple([self.count] + list(x.shape[1:])), dtype=v.dtype)
        for ell in range(0, self.ell_max + 1):
            k_max = self.k_max[ell]
            idx_radial = ind_radial + np.arange(0, k_max)
            # include the normalization factor of angular part
            ang_nrms = self.angular_norms[idx_radial]
            radial = self._precomp["radial"][:, idx_radial]
            radial = radial / ang_nrms

            sgns = (1,) if ell == 0 else (1, -1)
            for _ in sgns:
                ang = self._precomp["ang"][:, ind_ang]
                ang_radial = np.expand_dims(ang[ang_idx], axis=1) * radial[r_idx]
                idx = ind + np.arange(0, k_max)
                v[idx] = ang_radial.T @ x[mask]
                ind += len(idx)
                ind_ang += 1

            ind_radial += len(idx_radial)

        v = roll_dim(v, sz_roll)
        return v.T  # RCOPT

    def to_complex(self, coef):
        """
        Return complex valued representation of coefficients.
        This can be useful when comparing or implementing methods
        from literature.

        There is a corresponding method, to_real.

        :param coef: Coefficients from this basis.
        :return: Complex coefficent representation from this basis.
        """

        if coef.ndim == 1:
            coef = coef.reshape(1, -1)

        if coef.dtype not in (np.float64, np.float32):
            raise TypeError("coef provided to to_complex should be real.")

        # Pass through dtype precions, but check and warn if mismatched.
        dtype = complex_type(coef.dtype)
        if coef.dtype != self.dtype:
            logger.warning(
                f"coef dtype {coef.dtype} does not match precision of basis.dtype {self.dtype}, returning {dtype}."
            )

        # Return the same precision as coef
        imaginary = dtype(1j)

        ccoef = np.zeros((coef.shape[0], self.complex_count), dtype=dtype)

        ind = 0
        idx = np.arange(self.k_max[0], dtype=int)
        ind += np.size(idx)

        ccoef[:, idx] = coef[:, idx]

        for ell in range(1, self.ell_max + 1):
            idx = ind + np.arange(self.k_max[ell], dtype=int)
            ccoef[:, idx] = (
                coef[:, self._pos[idx]] - imaginary * coef[:, self._neg[idx]]
            ) / 2.0

            ind += np.size(idx)

        return ccoef

    def to_real(self, complex_coef):
        """
        Return real valued representation of complex coefficients.
        This can be useful when comparing or implementing methods
        from literature.

        There is a corresponding method, to_complex.

        :param complex_coef: Complex coefficients from this basis.
        :return: Real coefficent representation from this basis.
        """
        if complex_coef.ndim == 1:
            complex_coef = complex_coef.reshape(1, -1)

        if complex_coef.dtype not in (np.complex128, np.complex64):
            raise TypeError("coef provided to to_real should be complex.")

        # Pass through dtype precions, but check and warn if mismatched.
        dtype = real_type(complex_coef.dtype)
        if dtype != self.dtype:
            logger.warning(
                f"Complex coef dtype {complex_coef.dtype} does not match precision of basis.dtype {self.dtype}, returning {dtype}."
            )

        coef = np.zeros((complex_coef.shape[0], self.count), dtype=dtype)

        ind = 0
        idx = np.arange(self.k_max[0], dtype=int)
        ind += np.size(idx)
        ind_pos = ind

        coef[:, idx] = complex_coef[:, idx].real

        for ell in range(1, self.ell_max + 1):
            idx = ind + np.arange(self.k_max[ell], dtype=int)
            idx_pos = ind_pos + np.arange(self.k_max[ell], dtype=int)
            idx_neg = idx_pos + self.k_max[ell]

            c = complex_coef[:, idx]
            coef[:, idx_pos] = 2.0 * np.real(c)
            coef[:, idx_neg] = -2.0 * np.imag(c)

            ind += np.size(idx)
            ind_pos += 2 * self.k_max[ell]

        return coef

    def calculate_bispectrum(
        self, coef, flatten=False, filter_nonzero_freqs=False, freq_cutoff=None
    ):
        """
        Calculate bispectrum for a set of coefs in this basis.

        The Bispectum matrix is of shape:
            (count, count, unique_radial_indices)

        where count is the number of complex coefficients.

        :param coef: Coefficients representing a (single) image expanded in this basis.
        :param flatten: Optionally extract symmetric values (tril) and then flatten.
        :param freq_cutoff: Truncate (zero) high k frequecies above (int) value, defaults off (None).
        :return: Bispectum matrix (complex valued).
        """

        # Bispectrum implementation expects the complex representation of coefficients.
        complex_coef = self.to_complex(coef)

        return super().calculate_bispectrum(
            complex_coef,
            flatten=flatten,
            filter_nonzero_freqs=filter_nonzero_freqs,
            freq_cutoff=freq_cutoff,
        )



import logging
from collections import OrderedDict

import numpy as np

from aspire.basis import FFBBasis2D, SteerableBasis2D
from aspire.covariance import BatchedRotCov2D
from aspire.operators import BlkDiagMatrix
from aspire.utils import complex_type, fix_signs, real_type

logger = logging.getLogger(__name__)


class FSPCABasis(SteerableBasis2D):
    """
    A class for Fast Steerable Principal Component Analaysis basis.

    FSPCA is an extension to Fourier Bessel representations
    (provided asF BBasis2D/FFBBasis2D), which computes combinations of basis
    coefficients coresponding to the princicpal components of image(s)
    represented in the provided basis.

    The principal components are computed from eigen decomposition of the
    covariance matrix, and when evaluated into the real domain and reshaped form
    the set of `eigenimages`.

    The algorithm is described in the publication:
    Z. Zhao, Y. Shkolnisky, A. Singer, Fast Steerable Principal Component Analysis,
    IEEE Transactions on Computational Imaging, 2 (1), pp. 1-12 (2016).

    """

    def __init__(
        self, src, basis=None, noise_var=None, components=None, batch_size=512
    ):
        """

        :param src: Source instance
        :param basis: Optional Fourier Bessel Basis (usually FFBBasis2D)
        :param components: Optionally assign number of principal components
            to use for the FSPCA basis.
            Default value of `None` will use `self.basis.count`.
        :param noise_var: Optionally assign noise variance.
            Default value of `None` will estimate noise with WhiteNoiseEstimator.
            Use 0 when using clean images so cov2d skips applying noisy covar coeffs..
        :param batch_size: Batch size for computing basis coefficients.
            `batch_size` is also passed to BatchedRotCov2D.
        """

        self.src = src
        self.batch_size = batch_size

        # Automatically generate basis if needed.
        if basis is None:
            basis = FFBBasis2D((self.src.L,) * 2, dtype=self.src.dtype)
        self.basis = basis

        # Components are used for `compress` during `build`.
        self.components = components or self.basis.count
        self._check_components()

        # check/warn dtypes
        self.dtype = self.src.dtype
        if self.basis.dtype != self.dtype:
            logger.warning(
                f"basis.dtype {self.basis.dtype} does not match"
                f" source {self.src.dtype}, using {self.dtype}."
            )

        self.count = self.basis.count
        self.complex_count = self.basis.complex_count
        self.angular_indices = self.basis.angular_indices
        self.radial_indices = self.basis.radial_indices
        self.signs_indices = self.basis._indices["sgns"]
        self.complex_angular_indices = self.basis.complex_angular_indices
        self.complex_radial_indices = self.basis.complex_radial_indices

        self.complex_indices_map = self._get_complex_indices_map()
        assert (
            len(self.complex_indices_map) == self.complex_count
        ), f"{len(self.complex_indices_map)} != {self.complex_count}"

        # Map Cached SteerableBasis2D index maps
        self._zero_angular_inds = self.basis._zero_angular_inds
        self._pos_angular_inds = self.basis._pos_angular_inds
        self._neg_angular_inds = self.basis._neg_angular_inds

        self.noise_var = noise_var  # noise_var is handled during `build` call.

        self.build()

    def _check_components(self):
        """
        Check that our (compressed) count is not larger than our basis count.
        """
        if self.components > self.basis.count:
            raise ValueError(
                f"Provided components {self.components} > {self.basis.count} basis coefficients."
                "  Reduce components."
            )

    def _get_complex_indices_map(self):
        """
        Private method for building an ordered dictionary mapping
        every complex coefficients (ell, q) pair
        with the real coefficients' flattened array index.

        The ordering stores the first occurance of a complex
        coefficient, which is used in compression to select the
        first k complex components.

        Once we know the first k complex components, this mapping
        further provides the index of both +/- real coefficients.

        This is subtly different than taking the top 2*k real
        components' coefficients and converting to complex,
        which often does not yield k complex components.
        """

        complex_indices_map = OrderedDict()
        for i in range(self.count):
            ell = self.angular_indices[i]
            q = self.radial_indices[i]
            sgn = self.signs_indices[i]

            complex_indices_map.setdefault((ell, q), [None, None])
            if sgn == 1:
                complex_indices_map[(ell, q)][0] = i
            elif sgn == -1:
                complex_indices_map[(ell, q)][1] = i
            else:
                raise ValueError("sgn should be +-1")

        return complex_indices_map

    def build(self):
        """
        Computes the FSPCA basis.

        This may take some time for large image stacks.
        """

        if self.noise_var is None:
            from aspire.noise import WhiteNoiseEstimator

            logger.info("Estimating the noise of images.")
            self.noise_var = WhiteNoiseEstimator(self.src).estimate()
        logger.info(f"Setting noise_var={self.noise_var}")

        cov2d = BatchedRotCov2D(
            src=self.src, basis=self.basis, batch_size=self.batch_size
        )
        covar_opt = {
            "shrinker": "frobenius_norm",
            "verbose": 0,
            "max_iter": 250,
            "iter_callback": [],
            "store_iterates": False,
            "rel_tolerance": 1e-12,
            "precision": "float64",
            "preconditioner": "identity",
        }
        self.mean_coef_est = cov2d.get_mean()
        self.covar_coef_est = cov2d.get_covar(
            mean_coeff=self.mean_coef_est,
            noise_var=self.noise_var,
            covar_est_opt=covar_opt,
        )

        # Create the arrays to be packed by _compute_spca
        self.eigvals = np.zeros(self.basis.count, dtype=self.dtype)

        self.eigvecs = BlkDiagMatrix.empty(2 * self.basis.ell_max + 1, dtype=self.dtype)

        # Perform the PCA over batches, storing the compressed coefficients.
        self._compute_spca()

        #  Complete compression by mutating class
        self._compress()

    def _compute_spca(self):
        """
        Algorithm 2 from paper.

        It has been adopted to use ASPIRE-Python's
        cov2d (real) covariance estimation.
        """

        # -- Compute the spectrum blockwise. --
        # For each angular frequency (`ells` in FB code, `k` from paper)
        #   we use the properties of Block Diagonal Matrices to work
        #   on the correspong block.
        eigval_index = 0
        basis_inds = []
        for angular_index, C_k in enumerate(self.covar_coef_est):
            # # Eigen/SVD, covariance block C_k should be symmetric.
            eigvals_k, eigvecs_k = np.linalg.eigh(C_k)

            # Determistically enforce eigen vector sign convention
            eigvecs_k = fix_signs(eigvecs_k)

            # Sort eigvals_k
            sorted_indices = np.argsort(-eigvals_k)
            eigvals_k, eigvecs_k = (
                eigvals_k[sorted_indices],
                eigvecs_k[:, sorted_indices],
            )

            # These are the dense basis indices for this block.
            _basis_inds = np.arange(eigval_index, eigval_index + len(eigvals_k))
            basis_inds.append(_basis_inds)

            # Store the eigvals for this block, note this is a flat array.
            self.eigvals[_basis_inds] = eigvals_k

            # Store the eigvecs, note this is a BlkDiagMatrix and is assigned incrementally.
            self.eigvecs[angular_index] = eigvecs_k

            eigval_index += len(eigvals_k)

        # Sanity check we have same dimension of eigvals and basis coefs.
        if eigval_index != self.basis.count:
            raise RuntimeError(
                f"eigvals dimension {eigval_index} != basis coef count {self.basis.count}."
            )

        # Store a map of indices sorted by eigenvalue.
        #  We don't resort then now because this would destroy the block diagonal structure.
        #
        # sorted_indices[i] is the ith most powerful eigendecomposition index
        #
        # We can pass a full or truncated slice of sorted_indices to any array indexed by
        # the coefs.  This is used later for compression and index re-generation.
        self.sorted_indices = np.argsort(-np.abs(self.eigvals))

        compressed_indices = self._get_compressed_indices()

        self.spca_coef = np.zeros(
            (self.src.n, len(compressed_indices)), dtype=self.dtype
        )

        # Compute coefficient vector of mean image at zeroth component
        self.mean_coef_zero = self.mean_coef_est[self.angular_indices == 0]

        # Define mask for zero angular mode, used in loop below
        zero_ell_mask = self.basis._indices["ells"] == 0

        # Apply Data matrix batchwise
        num_batches = (self.src.n + self.batch_size - 1) // self.batch_size
        for i in range(num_batches):
            # Compute the coefficients for this batch
            start = i * self.batch_size
            finish = min((i + 1) * self.batch_size, self.src.n)
            batch_coef = self.basis.evaluate_t(self.src.images[start:finish])

            # Make the Data matrix (A_k)
            # # Construct A_k, matrix of expansion coefficients a^i_k_q
            # #   for image i, angular index k, radial index q,
            # #   (around eq 31-33)
            # #   Rows radial indices, columns image i.
            # #
            # # We can extract this directly (up to transpose) from
            # #  fb coef matrix where ells == angular_index
            # #  then use the transpose so image stack becomes columns.

            # Initialize a totally empty BlkDiagMatrix, then build incrementally.
            A = BlkDiagMatrix.empty(0, dtype=batch_coef.dtype)

            # Zero angular index is special case of indexing.
            A_0 = batch_coef[:, zero_ell_mask] - self.mean_coef_zero
            A.append(A_0)

            # Remaining angular indices have postive and negative entries in real representation.
            for ell in range(
                1, self.basis.ell_max + 1
            ):  # `ell` in this code is `k` from paper
                mask_ell = self.basis._indices["ells"] == ell
                mask_pos = mask_ell & (self.basis._indices["sgns"] == +1)
                mask_neg = mask_ell & (self.basis._indices["sgns"] == -1)

                A.append(batch_coef[:, mask_pos])
                A.append(batch_coef[:, mask_neg])

            if len(A) != len(self.covar_coef_est):
                raise RuntimeError(
                    "Data matrix A should have same number of blocks as Covar matrix.",
                    f" {len(A)} != {len(self.covar_coef_est)}",
                )

            # -- Compute new FSPCA coefficients. --
            # For each batch
            #   For each angular frequency (`ells` in FB code, `k` from paper)
            #     Use the properties of Block Diagonal Matrices to work
            #     on the correspong block.
            blk_spca_coef = np.empty_like(batch_coef)
            for angular_index, a_blk in enumerate(A):
                # To compute new expansion coefficients using spca basis
                #   we combine the basis coefs using the eigen decomposition.
                # Note image stack slow moving axis, otherwise this is just a
                #   block by block matrix multiply.
                blk_spca_coef[:, basis_inds[angular_index]] = (
                    a_blk @ self.eigvecs[angular_index]
                )

            # Assign truncated block to global spca_coef
            self.spca_coef[start:finish, :] = blk_spca_coef[:, compressed_indices]

    def expand_from_image_basis(self, x):
        """
        Take an image in the standard coordinate basis and express as FSPCA coefs.

        Note each FSPCA coef corresponds to a linear combination Fourier Bessel
        basis vectors, described by an eigenvector in FSPCA.

        :param x:  The Image instance representing a stack of images in the
            standard 2D coordinate basis to be evaluated.
        :return: Stack of coefs in the FSPCABasis.
        """
        fb_coefs = self.basis.evaluate_t(x)
        return self.expand(fb_coefs)

    def expand(self, x):
        """
        Take a Fourier-Bessel coefs and express as FSPCA coefs.

        Note each FSPCA coef corresponds to a linear combination Fourier Bessel
        basis vectors, described by an eigenvector in FSPCA.

        :param x:  Coefs representing a stack in the
            Fourier Bessel basis.
        :return: Stack of coefs in the FSPCABasis.
        """

        # Apply linear combination defined by FSPCA (eigvecs)
        c_fspca = x @ self.eigvecs

        assert c_fspca.shape == (x.shape[0], self.count)

        return c_fspca

    def evaluate_to_image_basis(self, c):
        """
        Take FSPCA coefs and evaluate as image in the standard coordinate basis.

        :param c:  Stack of coefs in the FSPCABasis to be evaluated.
        :return: The Image instance representing a stack of images in the
            standard 2D coordinate basis..
        """
        c_fb = self.evaluate(c)

        return self.basis.evaluate(c_fb)

    def evaluate(self, c):
        """
        Take FSPCA coefs and evaluate to Fourier Bessel (self.basis) ceofs.

        :param c:  Stack of coefs in the FSPCABasis to be evaluated.
        :return: The (real) coefs representing a stack of images in self.basis
        """

        # apply FSPCA eigenvector to coefs c, yields coefs in self.basis
        eigvecs = self.eigvecs
        if isinstance(eigvecs, BlkDiagMatrix):
            eigvecs = eigvecs.dense()

        # # XXX Legacy code doubles nonzero coefs ?
        # corrected_c = c.copy()
        # corrected_c[:, self.angular_indices!=0] *= 2
        # return corrected_c @ eigvecs.T

        return c @ eigvecs.T

    # TODO: Python>=3.8 @cached_property
    def _get_compressed_indices(self):
        """
        Return the sorted compressed (truncated) indices into the full FSPCA basis.

        Note that we return some number of indices in the real representation (in +- pairs)
        required to cover the `self.components` in the complex representation.
        """

        unsigned_components = zip(
            self.angular_indices[self.sorted_indices],
            self.radial_indices[self.sorted_indices],
        )

        # order the components by their importance (occurance based on sorted eigvals)
        #  This isn't exactly right since the eigvals would be sorted by the complex magnitude,
        #    instead of the larger component.
        # Note, that we only use OrderedDict for its key's (ie as an OrderedSet)
        ordered_components = OrderedDict()
        for k, q in unsigned_components:
            ordered_components.setdefault((k, q))  # inserts when not exists yet

        # Select the top n (k,q) pairs
        top_components = list(ordered_components)[: self.components]

        # Now we need to find the locations of both the + and - sgns.
        pos_mask = self.basis._indices["sgns"] == 1
        neg_mask = self.basis._indices["sgns"] == -1
        compressed_indices = []
        for k, q in top_components:
            # Compute the locations of coefs we're interested in.
            k_maps = self.angular_indices == k
            q_maps = self.radial_indices == q

            pos_index = np.argmax(k_maps & q_maps & pos_mask)
            compressed_indices.append(pos_index)
            if k > 0:
                neg_index = np.where(k_maps & q_maps & neg_mask)[0][0]
                compressed_indices.append(neg_index)
        return compressed_indices

    def _compress(self):
        """
        Use the eigendecomposition to select the most powerful
        coefficients.

        Using those coefficients new indice mappings are constructed.

        Mutates `self`.

        :param n: Number of components (coef)
        """

        if self.components >= self.count:
            logger.warning(
                f"Requested compression to {self.components} components,"
                f" but already {self.count}."
                "  Skipping compression."
            )
            return self

        # Create compressed mapping
        compressed_indices = self._get_compressed_indices()
        self.count = len(compressed_indices)

        self.eigvals = self.eigvals[compressed_indices]
        if isinstance(self.eigvecs, BlkDiagMatrix):
            self.eigvecs = self.eigvecs.dense()
        self.eigvecs = self.eigvecs[:, compressed_indices]

        self.angular_indices = self.angular_indices[compressed_indices]
        self.radial_indices = self.radial_indices[compressed_indices]
        self.signs_indices = self.signs_indices[compressed_indices]

        self.complex_indices_map = self._get_complex_indices_map()
        self.complex_count = len(self.complex_indices_map)
        self.complex_angular_indices = np.empty(self.complex_count, int)
        self.complex_radial_indices = np.empty(self.complex_count, int)
        for i, key in enumerate(self.complex_indices_map.keys()):
            ang, rad = key
            self.complex_angular_indices[i] = ang
            self.complex_radial_indices[i] = rad

    def to_complex(self, coef):
        """
        Return complex valued representation of coefficients.
        This can be useful when comparing or implementing methods
        from literature.

        There is a corresponding method, to_real.

        :param coef: Coefficients from this basis.
        :return: Complex coefficent representation from this basis.
        """

        if coef.ndim == 1:
            coef = coef.reshape(1, -1)

        if coef.dtype not in (np.float64, np.float32):
            raise TypeError("coef provided to to_complex should be real.")

        # Pass through dtype precions, but check and warn if mismatched.
        dtype = complex_type(coef.dtype)
        if coef.dtype != self.dtype:
            logger.warning(
                f"coef dtype {coef.dtype} does not match precision of basis.dtype {self.dtype}, returning {dtype}."
            )

        # Return the same precision as coef
        imaginary = dtype(1j)

        ccoef = np.zeros((coef.shape[0], self.complex_count), dtype=dtype)

        ccoef_d = OrderedDict()

        for i in range(self.count):
            ell = self.angular_indices[i]
            q = self.radial_indices[i]
            sgn = self.signs_indices[i]

            ccoef_d.setdefault((ell, q), 0 + 0j)
            if ell == 0:
                ccoef_d[(ell, q)] = coef[:, i]
            elif sgn == 1:
                ccoef_d[(ell, q)] += coef[:, i] / 2.0
            elif sgn == -1:
                ccoef_d[(ell, q)] -= imaginary * coef[:, i] / 2.0
            else:
                raise ValueError("sgns should be +-1")

        for i, k in enumerate(ccoef_d.keys()):
            ccoef[:, i] = ccoef_d[k]

        return ccoef

    def to_real(self, complex_coef):
        """
        Return real valued representation of complex coefficients.
        This can be useful when comparing or implementing methods
        from literature.

        There is a corresponding method, to_complex.

        :param complex_coef: Complex coefficients from this basis.
        :return: Real coefficent representation from this basis.
        """

        if complex_coef.ndim == 1:
            complex_coef = complex_coef.reshape(1, -1)

        if complex_coef.dtype not in (np.complex128, np.complex64):
            raise TypeError("coef provided to to_real should be complex.")

        # Pass through dtype precions, but check and warn if mismatched.
        dtype = real_type(complex_coef.dtype)
        if dtype != self.dtype:
            logger.warning(
                f"Complex coef dtype {complex_coef.dtype} does not match precision of basis.dtype {self.dtype}, returning {dtype}."
            )

        coef = np.zeros((complex_coef.shape[0], self.count), dtype=dtype)

        # map ordered index to (ell, q) key in dict
        keymap = list(self.complex_indices_map.keys())
        for i in range(self.complex_count):
            # retreive index into reals
            pos_i, neg_i = self.complex_indices_map[keymap[i]]
            if self.complex_angular_indices[i] == 0:
                coef[:, pos_i] = complex_coef[:, i].real
            else:
                coef[:, pos_i] = 2.0 * complex_coef[:, i].real
                coef[:, neg_i] = -2.0 * complex_coef[:, i].imag

        return coef

    def calculate_bispectrum(
        self, coef, flatten=False, filter_nonzero_freqs=False, freq_cutoff=None
    ):
        if coef.dtype == real_type(self.dtype):
            coef = self.to_complex(coef)
        return super().calculate_bispectrum(
            coef,
            flatten=flatten,
            filter_nonzero_freqs=filter_nonzero_freqs,
            freq_cutoff=freq_cutoff,
        )

    def eigen_images(self):
        """
        Return the eigen images of the FSPCA basis, evaluated to image space.

        This may be used to implot visualizations of the eigenvectors.

        Ordering corresponds to FSPCA eigvals.
        """

        eigvecs = self.eigvecs
        if isinstance(eigvecs, BlkDiagMatrix):
            eigvecs = eigvecs.dense()

        return self.basis.evaluate(eigvecs.T)

    def shift(self, coef, shifts):
        """
        Returns coefs shifted by `shifts`.

        This will transform to real cartesian space, shift,
        and transform back to Polar Fourier-Bessel space.

        :param coef: Basis coefs.
        :param shifts: Shifts in pixels (x,y). Shape (1,2) or (len(coef), 2).
        :return: coefs of shifted images.
        """

        shifts = np.atleast_2d(np.array(shifts))
        if shifts.ndim != 2:
            raise ValueError("`shifts` should be a one or two dimensional array.")
        if shifts.shape[1] != 2 or shifts.shape[0] not in (1, len(coef)):
            raise ValueError(
                "`shifts` should be shape (1,2) or (len(coef),2),"
                f" received {shifts.shape}."
            )

        # This transforms FSPCA->FB->Image->FB->FSPCA
        return self.expand_from_image_basis(
            self.evaluate_to_image_basis(coef).shift(shifts)
        )
